#!/usr/bin/env python3
"""
PubChem/PUG-SOAP sub|sim|exact search.
Originally pug_search.py (~2010)
https://pubchem.ncbi.nlm.nih.gov/pug_soap/pug_soap.cgi?wsdl
"""
import sys,os,re,urllib,time,argparse,tempfile,logging

from ... import pubchem
from ...pubchem.soap import Utils

#############################################################################
def ElapsedTime(t0):
  return (time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))

#############################################################################
if __name__=='__main__':
  PUGURL='https://pubchem.ncbi.nlm.nih.gov/pug/pug.cgi';
  POLL_WAIT=10; MAX_WAIT=300;
  SIM_CUTOFF=0.80;
  NMAX=100;
  SEARCHTYPES = ["substructure", "similarity", "exact", "standardize"]

  parser = argparse.ArgumentParser(description="PubChem PUG SOAP client: sub|sim|exact search, fetch smiles|sdfs")
  parser.add_argument("--o", dest="ofile", help="output smiles|sdf file (w/ CIDs)")
  parser.add_argument("--qsmi", required=True, help="input query smiles (or smarts)")
  parser.add_argument("--ofmt", choices=["smiles", "sdf"], help="output format [or via filename]")
  parser.add_argument("--gz", action="store_true", help="output gzipped [default via filename]")
  parser.add_argument("--out_cids", action="store_true", help="output CIDs (only w/ smiles output)")
  parser.add_argument("--searchtype", choices=SEARCHTYPES, default="exact", help="search type")
  parser.add_argument("--sim_cutoff", type=float, default=SIM_CUTOFF, help="similarity cutoff, Tanimoto")
  parser.add_argument("--active", help="active mols only (in any assay)")
  parser.add_argument("--mlsmr", help="MLSMR mols only")
  parser.add_argument("--nmax", type=int, default=NMAX, help="max hits returned")
  parser.add_argument("--max_wait", type=int, default=MAX_WAIT, help="max wait for query")
  parser.add_argument("--poll_wait", type=int, default=POLL_WAIT, help="polling wait interval")
  parser.add_argument("--pugurl", default=PUGURL)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  logging.info('Query: "{}"'.format(args.qsmi))

  ofmt = args.ofmt if args.ofmt else None

  do_gz = args.gz if args.gz else False

  fout = open(args.ofile, 'wb') if args.ofile else sys.stdout

  if args.ofile:
    if (args.ofile[-3:]=='.gz'):
      do_gz = True
      ext = re.sub('^.*\.', '', args.ofile[:-3])
    else:
      ext = re.sub('^.*\.', '', args.ofile)
    if not args.ofmt: ofmt = ext

  if not ofmt: ofmt = "smiles"
  elif ofmt[:3].lower()=='smi': ofmt='smiles'
  elif ofmt[:3].lower() in ('sdf', 'mol', 'mdl'): ofmt='sdf'

  t0=time.time()	## start timer

  if args.searchtype == "standardize":
    pugreq = pubchem.soap.Utils.SubmitQuery_Standardize(args.pugurl, args.qsmi)
  else:
    pugreq = pubchem.soap.Utils.SubmitQuery_Structural(args.pugurl, args.qsmi, args.searchtype, args.sim_cutoff, args.active, args.mlsmr, args.nmax)

  ### error status values include: "input-error", "server-error"
  if pugreq.status not in ("success", "queued", "running"):
    logging.error("""Query failed; quitting (status={}, error="{}").""".format(pugreq.status, pugreq.error))
    sys.exit()
  logging.info("Query [{}] status: {}".format(pugreq.reqid, pugreq.status))

  while pugreq.status != "success": 
    time.sleep(args.poll_wait)
    if time.time()-t0>args.max_wait:
      pugreq.cancel()
      logging.error('Max wait exceeded ({} sec); quitting.'.format(args.max_wait))
    logging.info('Polling PUG [{}]...'.format(pugreq.reqid))
    pugreq.getStatus()
    if pugreq.qkey: break
    logging.info('{} elapsed; {} sec wait; status={}'.format(ElapsedTime(t0), args.poll_wait, pugreq.status))
    if pugreq.error: logging.info(pugreq.error)
    if pugreq.error and re.search('No result found', pugreq.error, re.I):
      sys.exit()

  if args.searchtype == "standardize":
    fout.write(pugreq.out_smi+"\n")
    sys.exit()

  pugreq2 = pubchem.soap.Utils.MonitorQuery(args.pugurl, pugreq.qkey, pugreq.wenv, ofmt, do_gz)
  while True:
    if time.time()-t0>args.max_wait:
      pugreq2.cancel()
      logging.error('Max wait exceeded ({} sec); quitting.'.format(args.max_wait))
    logging.info('Polling PUG [{}]...'.format(pugreq2.reqid))
    pugreq2.getStatus()
    if pugreq2.url: break
    logging.info('{} elapsed; {} sec wait; status={}'.format(ElapsedTime(t0), args.poll_wait, pugreq2.status))
    time.sleep(args.poll_wait)

  pubchem.soap.Utils.DownloadResults(pugreq2.url, fout)

  if args.ofile:
    logging.info('URL: {} downloaded to {} ({:.2f}MB)'.format(pugreq2.url, args.ofile, os.stat(args.ofile).st_size/1e6))
  fout.close()
  logging.info('Format: {}{}'.format(ofmt, (' (gzipped)' if do_gz else '')))

  logging.info('Elapsed time: {}'.format(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python3
"""
TIN-X (Target Importance and Novelty Explorer) REST API client
https://www.newdrugtargets.org/
https://api.newdrugtargets.org/docs
https://api.newdrugtargets.org/targets/
"""
###
import sys,os,re,argparse,time,logging
#
from .. import tinx
#
API_HOST="api.newdrugtargets.org"
API_BASE_PATH=""
#
##############################################################################
if __name__=='__main__':
  epilog="""\
Example IDs:
5391 (disease); DOID:9297 (DOID);
12203 (target); Q9H4B4 (UniProt);
--disease_ids needed ONLY if BOTH target and disease IDs specified, such
as for get_disease_target_articles.
"""
  parser = argparse.ArgumentParser(description="TIN-X (Target Importance and Novelty Explorer) REST API client)", epilog=epilog)
  ops = [
	"list_diseases",
	"list_targets",
	"list_articles",
	"list_dto",
	"get_disease",
	"get_disease_by_doid",
	"get_disease_targets",
	"get_disease_target_articles",
	"get_target",
	"get_target_by_uniprot",
	"get_target_diseases",
	"get_dto",
	"search_diseases",
	"search_targets",
	"search_articles",
	"search_dtos"
	]
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--i", dest="ifile", help="input IDs or search terms")
  parser.add_argument("--ids", help="IDs or search terms (comma-separated)")
  parser.add_argument("--disease_ids", help="disease IDs (comma-separated), needed ONLY if BOTH target and disease IDs specified")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="search query")
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if ids: logging.info('Input IDs: %d'%(len(ids)))
  if args.disease_ids:
    disease_ids = re.split('[, ]+', args.disease_ids.strip())

  t0=time.time()

  base_url = 'https://'+args.api_host+args.api_base_path

  if args.op=='list_targets':
    tinx.Utils.ListTargets(base_url, args.skip, args.nmax, fout)

  elif args.op=='list_diseases':
    tinx.Utils.ListDiseases(base_url, args.skip, args.nmax, fout)

  elif args.op=='list_articles':
    tinx.Utils.ListArticles(base_url, args.skip, args.nmax, fout)

  elif args.op=='list_dto':
    tinx.Utils.ListDTO(base_url, args.skip, args.nmax, fout)

  elif args.op=='get_disease':
    tinx.Utils.GetDisease(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='get_disease_by_doid':
    tinx.Utils.GetDiseaseByDOId(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='get_disease_targets':
    tinx.Utils.GetDiseaseTargets(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='get_disease_target_articles':
    tinx.Utils.GetDiseaseTargetArticles(base_url, disease_ids, ids, args.skip, args.nmax, fout)

  elif args.op=='get_target':
    tinx.Utils.GetTarget(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='get_target_by_uniprot':
    tinx.Utils.GetTargetByUniprot(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='get_target_diseases':
    tinx.Utils.GetTargetDiseases(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='search_diseases':
    tinx.Utils.SearchDiseases(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='search_targets':
    tinx.Utils.SearchTargets(base_url, ids, args.skip, args.nmax, fout)

  elif args.op=='search_articles':
    tinx.Utils.SearchArticles(base_url, ids, args.skip, args.nmax, fout)

  else:
    parser.error('Unknown operation: %s'%args.op)

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
"""
LINCS REST API client
New (2019) iLINCS: 
http://www.ilincs.org/ilincs/APIinfo
http://www.ilincs.org/ilincs/APIdocumentation
(http://lincsportal.ccs.miami.edu/dcic/api/ DEPRECATED?)
"""
###
import sys,os,re,json,logging
#
from ..util import rest_utils
#
#############################################################################
def GetGene(base_url, ids, fout):
  tags=None;
  for id_this in ids:
    url = base_url+'/GeneInfos/%s'%id_this
    rval = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2))
    if not tags:
      tags = [tag for tag in rval.keys() if type(rval[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(rval[tag]) if tag in rval else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("IDs: {0}".format(len(ids)))

#############################################################################
def GetDataset(base_url, ids, fout):
  tags=None;
  for id_this in ids:
    url = base_url+'/PublicDatasets/%s'%id_this
    rval = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2))
    if not tags:
      tags = [tag for tag in rval.keys() if type(rval[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(rval[tag]) if tag in rval else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("IDs: {0}".format(len(ids)))

#############################################################################
def GetCompound(base_url, ids, fout):
  tags=None; n_cpd=0;
  for id_this in ids:
    url = base_url+'/Compounds/%s'%id_this
    cpd = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(cpd, indent=2))
    if not tags:
      tags = [tag for tag in cpd.keys() if type(cpd[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(cpd[tag]) if tag in cpd else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_cpd+=1
  logging.info("IDs: {0}; n_cpd: {1}".format(len(ids), n_cpd))

#############################################################################
def SearchDataset(base_url, searchTerm, lincs, fout):
  tags=None;
  url = base_url+'/PublicDatasets/findTermMeta'
  d = {'term':searchTerm}
  if lincs: d['lincs'] = True
  rval = rest_utils.PostURL(url, data=d, parse_json=True)
  logging.debug(json.dumps(rval, indent=2))
  dsets = rval['data'] if 'data' in rval else []
  for dset in dsets:
    logging.debug(json.dumps(dset, indent=2))
    if not tags:
      tags = [tag for tag in dset.keys() if type(dset[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(dset[tag]) if tag in dset else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("Datasets: {0}".format(len(dsets)))

#############################################################################
def SearchSignature(base_url, ids, lincs, fout):
  #SignatureMeta?filter={"where":{"lincspertid":"LSM-2121"},"limit":10}
  tags=None; n_sig=0;
  for id_this in ids:
    url = base_url+'/SignatureMeta?filter={"where":{"lincspertid":"%s"}}'%id_this
    sigs = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(sigs, indent=2))
    for sig in sigs:
      logging.debug(json.dumps(sig, indent=2))
      if not tags:
        tags = [tag for tag in sig.keys() if type(sig[tag]) not in (list, dict)]
        fout.write('\t'.join(tags)+'\n')
      vals = [str(sig[tag]) if tag in sig else '' for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_sig+=1
  logging.info("IDs: {0}; n_sig: {1}".format(len(ids), n_sig))

#############################################################################
def GetSignature(base_url, ids, ngene, fout):
  tags=None; n_gene=0;
  url = base_url+'/ilincsR/downloadSignature'
  d = {'sigID':(','.join(ids)), 'display':True, 'noOfTopGenes':ngene}
  rval = rest_utils.PostURL(url, data=d, parse_json=True)
  logging.debug(json.dumps(rval, indent=2))
  genes = rval['data']['signature'] if 'data' in rval and 'signature' in rval['data'] else []
  for gene in genes:
    logging.debug(json.dumps(gene, indent=2))
    if not tags:
      tags = [tag for tag in gene.keys() if type(gene[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(gene[tag]) if tag in gene else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_gene+=1
  logging.info("IDs: {0}; n_gene: {1}".format(len(ids), n_gene))

#############################################################################



#!/usr/bin/env python3
###
import sys,os,json,re,logging
#
from ..util import rest_utils

#############################################################################
def GetStats(base_url, pmids, fout):
  """Request multiply by chunk. Lists of PMIDs (e.g. references,
cited_by) reported as counts."""
  n_in=0; n_out=0; tags=None; nchunk=100;
  while True:
    if n_in>=len(pmids): break
    pmids_this = pmids[n_in:n_in+nchunk]
    n_in += (nchunk if n_in+nchunk < len(pmids) else len(pmids)-n_in)
    url_this = (base_url+'?pmids=%s'%(','.join(pmids_this)))
    rval = rest_utils.GetURL(url_this, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, indent=2))
    url_self = rval['links']['self']
    pubs = rval['data']
    for pub in pubs:
      if not tags:
        tags = list(pub.keys())
        fout.write('\t'.join(tags)+'\n')
      vals=[];
      for tag in tags:
        val = (pub[tag] if tag in pub else '')
        if type(val) is list: val = len(val)
        elif type(val) is float: val = "{:.3f}".format(val)
        vals.append(val)
      fout.write('\t'.join([str(val) for val in vals])+'\n')
      n_out+=1
  logging.info('n_in: %d; n_out: %d'%(len(pmids), n_out))

#############################################################################
def GetStats_single(base_url, pmids, fout):
  """Request singly."""
  n_out=0; tags=None;
  for pmid in pmids:
    url = base_url+'/%s'%pmid
    pub = rest_utils.GetURL(url, parse_json=True)
    if not pub:
      logging.info('not found: %s'%(pmid))
      continue
    if not tags:
      tags = pub.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(pub[tag]) if tag in pub else '') for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('n_in: %d; n_out: %d'%(len(pmids), n_out))

#############################################################################



#!/usr/bin/env python3
"""
Utility functions for PUG-SOAP API (pre-REST) (~2010).
https://pubchemdocs.ncbi.nlm.nih.gov/pug-soap
See Save Query button for template XML at
https://pubchem.ncbi.nlm.nih.gov/search/search.cgi
And at
https://pubchem.ncbi.nlm.nih.gov/standardize/standardize.cgi
"""
import sys,os,io,re,urllib.request,time,logging

from xml.etree import ElementTree
#from xml.parsers import expat

from ...util import rest_utils
from ...util import xml_utils

XMLHEADER=("""\
<?xml version="1.0"?>
<!DOCTYPE PCT-Data PUBLIC "-//NCBI//NCBI PCTools/EN" "https://pubchem.ncbi.nlm.nih.gov/pug/pug.dtd">
""")

SOAP_ACTIONS = {
	"exact":"http://pubchem.ncbi.nlm.nih.gov/SubstructureSearch",
	"substructure":"http://pubchem.ncbi.nlm.nih.gov/SubstructureSearch",
	"similarity":"http://pubchem.ncbi.nlm.nih.gov/SimilaritySearch2D",
	"standardize":"http://pubchem.ncbi.nlm.nih.gov/Standardize"
	}

#############################################################################
class PugSoapRequest:
  """
Normal status values: "success", "queued", "running"
Error status values: "server-error"
  """

  def __init__(self, pugurl, pugxml):

    self.pugurl = pugurl
    self.pugxml = pugxml
    self.reqid = None
    self.status = None
    self.url = None
    self.qkey = None
    self.wenv = None
    self.error = None
    self.ntries = None
    self.out_smi = None
    self.parsePugXml(pugxml)
    self.qxml_template="""\
<PCT-Data><PCT-Data_input>
  <PCT-InputData><PCT-InputData_request>
    <PCT-Request>
      <PCT-Request_reqid>{REQID}</PCT-Request_reqid>
      <PCT-Request_type value="{MODE}"/>
    </PCT-Request>
  </PCT-InputData_request></PCT-InputData>
</PCT-Data_input></PCT-Data>
"""

  def getStatus(self):
    logging.debug("Connecting {}...".format(self.pugurl))
    pugxml = rest_utils.PostURL(self.pugurl, data=(XMLHEADER+self.qxml_template.format(REQID=self.reqid, MODE="status")), headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":"http://pubchem.ncbi.nlm.nih.gov/GetOperationStatus"})
    self.parsePugXml(pugxml)

  def cancel(self):
    logging.debug("Connecting {}...".format(self.pugurl))
    pugxml = rest_utils.PostURL(self.pugurl, data=(XMLHEADER+qxml_template.format(REQID=self.reqid, MODE="cancel")), headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":"http://pubchem.ncbi.nlm.nih.gov/GetOperationStatus"})
    self.parsePugXml(pugxml)

  def parsePugXml(self, pugxml):
    logging.debug(pugxml)
    try:
      elt = ElementTree.parse(io.StringIO(pugxml)) #ElementTree instance
    except Exception as e:
      logging.info('XML parse error: {}; xml="{}"'.format(e, pugxml))
      sys.exit(1)
    statuss = xml_utils.DOM_GetAttr(elt, 'PCT-Status', 'value')
    self.status = statuss[0] if statuss else None
    errors = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Status-Message_message')
    self.error = errors[0] if errors else None
    reqids = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Waiting_reqid')
    self.reqid = int(reqids[0]) if reqids else None
    urls = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Download-URL_url')
    self.url=urls[0] if urls else None
    qkeys = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Entrez_query-key')
    self.qkey = qkeys[0] if qkeys else None
    wenvs = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Entrez_webenv')
    self.wenv = wenvs[0] if wenvs else None
    #Standardize:
    out_smis = xml_utils.DOM_GetLeafValsByTagName(elt, 'PCT-Structure_structure_string')
    self.out_smi = out_smis[0] if out_smis else None

#############################################################################
def SubmitQuery_Structural(pugurl, smiles, searchtype, sim_cutoff, active=False, mlsmr=False, nmax=10000):
  qxml="""\
<PCT-Data><PCT-Data_input><PCT-InputData><PCT-InputData_query><PCT-Query>
  <PCT-Query_type><PCT-QueryType><PCT-QueryType_css>
    <PCT-QueryCompoundCS>
    <PCT-QueryCompoundCS_query>
      <PCT-QueryCompoundCS_query_data>{SMILES}</PCT-QueryCompoundCS_query_data>
    </PCT-QueryCompoundCS_query>
    <PCT-QueryCompoundCS_type>
""".format(SMILES=smiles)
  if searchtype.lower()=='similarity':
    qxml+="""\
<PCT-QueryCompoundCS_type_similar><PCT-CSSimilarity>
  <PCT-CSSimilarity_threshold>{SIM_CUTOFF:d}</PCT-CSSimilarity_threshold>
</PCT-CSSimilarity></PCT-QueryCompoundCS_type_similar>
""".format(SIM_CUTOFF=(int(sim_cutoff*100)))
  elif searchtype.lower()=='substructure':
    qxml+="""\
<PCT-QueryCompoundCS_type_subss><PCT-CSStructure>
  <PCT-CSStructure_isotopes value="false"/>
  <PCT-CSStructure_charges value="false"/>
  <PCT-CSStructure_tautomers value="false"/>
  <PCT-CSStructure_rings value="false"/>
  <PCT-CSStructure_bonds value="true"/>
  <PCT-CSStructure_chains value="true"/>
  <PCT-CSStructure_hydrogen value="false"/>
</PCT-CSStructure></PCT-QueryCompoundCS_type_subss>
"""
  else: #exact
    qxml+="""\
<PCT-QueryCompoundCS_type_identical>
  <PCT-CSIdentity value="same-stereo-isotope">5</PCT-CSIdentity>
</PCT-QueryCompoundCS_type_identical>
"""
  qxml+="""\
</PCT-QueryCompoundCS_type>
<PCT-QueryCompoundCS_results>{NMAX}</PCT-QueryCompoundCS_results>
</PCT-QueryCompoundCS>
</PCT-QueryType_css></PCT-QueryType>
""".format(NMAX=nmax)
  if active:
    qxml+="""\
<PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
<PCT-QueryCompoundEL_activity>
<PCT-CLByBioActivity value="active">2</PCT-CLByBioActivity>
</PCT-QueryCompoundEL_activity>
</PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
"""
  if mlsmr:
    qxml+="""\
<PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
<PCT-QueryCompoundEL_source><PCT-CLByString>
<PCT-CLByString_qualifier value="must">1</PCT-CLByString_qualifier>
<PCT-CLByString_category>MLSMR</PCT-CLByString_category>
</PCT-CLByString></PCT-QueryCompoundEL_source>
</PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
"""
  qxml+="""\
</PCT-Query_type></PCT-Query>
</PCT-InputData_query></PCT-InputData></PCT-Data_input></PCT-Data>
"""
  logging.debug("Requesting: pugurl: {}; searchtype: {}; smiles: {}".format(pugurl, searchtype, smiles))
  if searchtype.lower()=='similarity':
    logging.debug("Similarity cutoff: {}".format(sim_cutoff))
  pugxml = rest_utils.PostURL(pugurl, data=(XMLHEADER+qxml), headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":SOAP_ACTIONS[searchtype]})
  return PugSoapRequest(pugurl, pugxml)

#############################################################################
def SubmitQuery_Standardize(pugurl, smiles):
  qxml="""\
<PCT-Data>
  <PCT-Data_input>
    <PCT-InputData>
      <PCT-InputData_standardize>
        <PCT-Standardize>
          <PCT-Standardize_structure>
            <PCT-Structure>
              <PCT-Structure_structure>
                <PCT-Structure_structure_string>{SMILES}</PCT-Structure_structure_string>
              </PCT-Structure_structure>
              <PCT-Structure_format>
                <PCT-StructureFormat value="smiles"/>
              </PCT-Structure_format>
            </PCT-Structure>
          </PCT-Standardize_structure>
          <PCT-Standardize_oformat>
            <PCT-StructureFormat value="smiles"/>
          </PCT-Standardize_oformat>
        </PCT-Standardize>
      </PCT-InputData_standardize>
    </PCT-InputData>
  </PCT-Data_input>
</PCT-Data>
""".format(SMILES=smiles)
  logging.debug("Requesting: pugurl: {}; smiles: {}".format(pugurl, smiles))
  pugxml = rest_utils.PostURL(pugurl, data=(XMLHEADER+qxml), headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":SOAP_ACTIONS["standardize"]})
  return PugSoapRequest(pugurl, pugxml)

#############################################################################
def MonitorQuery(pugurl, qkey, webenv, fmt='smiles', do_gz=False):
  qxml="""\
<PCT-Data><PCT-Data_input><PCT-InputData>
  <PCT-InputData_download><PCT-Download> 
    <PCT-Download_uids><PCT-QueryUids>
      <PCT-QueryUids_entrez><PCT-Entrez>
        <PCT-Entrez_db>pccompound</PCT-Entrez_db>
        <PCT-Entrez_query-key>{QKEY}</PCT-Entrez_query-key>
        <PCT-Entrez_webenv>{WEBENV}</PCT-Entrez_webenv>
      </PCT-Entrez></PCT-QueryUids_entrez>
    </PCT-QueryUids></PCT-Download_uids>
    <PCT-Download_format value="{FMT}"/>
    <PCT-Download_compression value="{COMPRES}"/>
  </PCT-Download></PCT-InputData_download>
</PCT-InputData></PCT-Data_input></PCT-Data>
""".format(QKEY=qkey, WEBENV=webenv, FMT=fmt, COMPRES=('gzip' if do_gz else 'none'))
  logging.info('Requesting: pugurl: {}; qkey: {}'.format(pugurl, qkey))
  pugxml = rest_utils.PostURL(pugurl, data=(XMLHEADER+qxml), headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":"http://pubchem.ncbi.nlm.nih.gov/Download"})
  return PugSoapRequest(pugurl, pugxml)

#############################################################################
def DownloadResults(pugurl, fout, ntries=10):
  pugout = rest_utils.GetURL(pugurl, nmax_retry=ntries, headers={"Accept":"text/soap+xml; charset=utf-8", "SOAPAction":"http://pubchem.ncbi.nlm.nih.gov/Download"})
  if fout == sys.stdout:
    fout.write(pugout)
  elif type(pugout) is bytes:
    fout.write(pugout)
  elif type(pugout) is str:
    fout.write(pugout.encode("utf-8"))
  else: # ?
    fout.write(pugout)

#############################################################################



#!/usr/bin/env python
"""BROKEN! TO DO: Port to Python3. Preferrably avoid R and switch to Numpy, Scipy."""
#############################################################################
### Use R via RPy2 to calculate Gini Index using ineq package.
#############################################################################
# rpy2 prerequisites:
#    - Python 2.4+
#    - R v2.7+, built with ./configure --enable-R-shlib
#    - Set R_HOME for Apache
#############################################################################
import os,sys,getopt,math,csv

#import rpy2.robjects

#############################################################################
def GiniRCode(vals):
  xs=[]
  for x in vals:
    if x==None: continue
    xs.append('%.4e'%x)
  if not xs:
    return 'gini<-0.0\n'
  
  r_code='''\
vals<-c(%(SCORES)s)
library(ineq)
gini<-Gini(vals)
'''%{'SCORES':(','.join(xs))}

  return r_code

#############################################################################
def LorentzPlotRCode(vals,ifile):
  r_code='''\
library(ineq)
'''
  xs=[]
  for x in vals:
    if x==None: continue
    xs.append('%.4e'%x)
  r_code+='vals<-c(%s)\n'%(','.join(xs))
  r_code+='''\
gini<-Gini(vals)
Lc.p <- Lc(vals)

#postscript()
pdf('%(FILE)s')
#png()
#jpeg()
#X11()

plot(Lc.p, col=2,
  main=paste('Lorenz Curve\nGini Index:',gini),
  xlab=paste('(N =',%(N)d,')'),
  ylab="score")

dev.off()
#system("acroread \"%(FILE)s\"")
'''%{'FILE':ifile,'N':len(vals)}

  return r_code

#############################################################################
def GiniProcessor(vals):
  r_code=GiniRCode(vals)
  rpy2.robjects.r(r_code)
  try:
    gini=rpy2.robjects.r('gini')[0]
  except:
    gini = 0.0
    print >>sys.stderr, "Error: R failure"
  return gini

#############################################################################
def GiniInteractive(vals,ifile):
  r_code=LorentzPlotRCode(vals,ifile)
  print >>sys.stderr, "DEBUG:\n", r_code
  rpy2.robjects.r(r_code)
  try:
    gini=rpy2.robjects.r('gini')[0]
  except:
    gini = 0.0
    print >>sys.stderr, "Error: R failure"
  return gini

#############################################################################
def Lorenz_img(vals,ofile,title,format='PNG',height=1000,width=1200,
	iconic=False):
  xs=[]
  while None in vals: vals.remove(None)
  if min(vals)==0.0 and max(vals)==0.0:
    return 'gini<-0.0\n'
  for x in vals:
    xs.append('%.4e'%x)
  if not xs:
    return 'gini<-0.0\n'

  r_code='''\
vals<-c(%(SCORES)s)
library(ineq)
gini<-Gini(vals)
Lc.p <- Lc(vals)
'''%{'SCORES':(','.join(xs))}

  if format.upper()=='EPS':
    r_code+='''\
postscript("%(OUTFILE)s",horizontal=FALSE,onefile=FALSE,paper="special",
        height=5,width=7,family="Bookman")
'''%{'OUTFILE':ofile}
  else:
    r_code+='''\
png(filename="%(OUTFILE)s",height=%(H)d,width=%(W)d,units="px")
'''%{'OUTFILE':ofile,'H':height,'W':width}

  if iconic:
    r_code+='''\
#plot(Lc.p,col=2,lwd=4,main="",xlab="",ylab="")
plot(Lc.p,col=2,lwd=4,main="",xlab="",ylab="",
	mar=c(0,0,0,0),
	bty="o",
	tck=0,
	tcl=0,
	pty="m",
	cex=0,
	xaxt="n",
	yaxt="n")
'''
  else:
    r_code+='''\
plot(Lc.p,col=2,lwd=8,main="",xlab="",ylab="")
title(main=paste("%(TITLE)s",'\nLorenz curve'),
        xlab=paste("Lorenz curve","\nN = %(N)d","\ngini = ",gini),ylab="score")
'''%{'TITLE':title,'N':len(vals)}

  r_code+='dev.off()\n'

  rpy2.robjects.r(r_code)

  try:
    gini=rpy2.robjects.r('gini')[0]
  except:
    print >>sys.stderr, "Error: R failure"
    gini = 0.0
  return gini

#############################################################################
def CSVFile2Values(fin,tag,delim=',',verbose=0):
  vals=[]
  n_in=0; n_out=0;
  csvReader=csv.DictReader(fin,fieldnames=None,restkey=None,restval=None,dialect='excel',delimiter=delim,quotechar='"')

  while True:
    try:
      csvrow=csvReader.next()
      n_in+=1
    except Exception, e:
      if str(e)!='':
        print >>sys.stderr, 'Error (Exception): %s'%e
      break

    if n_in==1 and tag not in csvReader.fieldnames:
      print >>sys.stderr, ('ERROR: cannot find field "%s" in ifile.'%tag)
      return vals

    val=csvrow[tag]
    if val=='': continue

    try:
      val=float(val)
      vals.append(val)
      n_out+=1
    except Exception, e:
      print >>sys.stderr, 'Error (Exception): %s'%e

    #if n_out>100000: break

  if verbose:
    print >>sys.stderr, 'n_in: %d ; n_out: %d'%(n_in,len(vals))

  return vals

#############################################################################
def Gini(fin,tag,delim=',',verbose=0):

  vals = CSVFile2Values(fin,tag,delim,verbose)
  g = GiniProcessor(vals)
  print 'DEBUG: gini=%.3f'%g

#############################################################################
def Test():
  valses=[]
  valses.append([0.0,0.5,1.0,0.5,0.0,1.0,1.0,1.0])
  valses.append([0.0,0.0,0.0,0.0,0.0,0.0,0.0,1.0])
  valses.append([1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0])

  
  for vals in valses:

    gini=GiniProcessor(vals)
    print 'DEBUG: gini=%.3f'%gini

    #gini=GiniInteractive(vals,"z.pdf")
    #os.system("acroread z.pdf")


#############################################################################
def Help(msg=''):
  if msg: print msg
  print '''\
%(PROG)s
required (one of):
        --gini .................. Gini coefficient for given input file, CSV tag
        --test
options:
        --i IFILE ............... CSV
        --o OFILE ............... 
        --tag TAG ............... CSV tag
        --v ..................... verbose
        --h ..................... this help
'''%{'PROG':PROG}
  sys.exit()

#############################################################################
if __name__=='__main__':
  def ErrorExit(msg): print >>sys.stderr,msg ; sys.exit(1)

  PROG=os.path.basename(sys.argv[0])

  ifile=''; ofile='';
  test=False;
  gini=False;
  tag=''; delim='\t';
  verbose=0;

  opts,pargs=getopt.getopt(sys.argv[1:],'',[
    'i=',
    'o=',
    'test',
    'gini',
    'tag=',
    'delim=',
    'h=','help','v','vv'])
  if not opts: Help()
  for (opt,val) in opts:
    if opt=='--help': Help()
    elif opt=='--i': ifile=val
    elif opt=='--o': ofile=val
    elif opt=='--test': test=True
    elif opt=='--gini': gini=True
    elif opt=='--tag': tag=val
    elif opt=='--delim': delim=val
    elif opt=='--v': verbose=1
    elif opt=='--vv': verbose=2
    else: Help('Illegal option: %s'%(opt))

  if ofile:
    fout=open(ofile,'w+')
    if not fout:
      ErrorExit('Could not open output file: %s'%ofile)
  else:
    fout=sys.stdout

  if ifile:
    fin=open(ifile)
    if not fin:
      ErrorExit('Could not open input file: %s'%ifile)

  if test:
    Test()
  elif gini:
    Gini(fin,tag,delim,verbose)
  else:
    ErrorExit('No action specified.')

  if ofile:
    fout.close()



#!/usr/bin/env python3
'''
	XML utility functions.
'''
import sys,os,re,argparse,logging

from xml.etree import ElementTree #newer
from xml.parsers import expat


#############################################################################
def DOM_NodeText(node):
  #logging.info('DEBUG: type(node) = %s'%str(type(node)))
  if type(node) in (ElementTree.Element, ElementTree.ElementTree):
    return DOM_NodeText_ET(node)
  else:
    return DOM_NodeText_minidom(node)

#############################################################################
def DOM_NodeText_ET(node):
  return node.text

#############################################################################
def DOM_NodeText_minidom(node):
  for cnode in node.childNodes:
    #if cnode.nodeName=='#text':
    if cnode.nodeType == cnode.TEXT_NODE:
      return cnode.nodeValue.strip()
  return ''

#############################################################################
def DOM_GetLeafValsByTagName(root, tag):
  if type(root) in (ElementTree.Element, ElementTree.ElementTree):
    return DOM_GetLeafValsByTagName_ET(root, tag)
  else:
    return DOM_GetLeafValsByTagName_minidom(root, tag)

#############################################################################
def DOM_GetLeafValsByTagName_ET(root, tag):
  vals=[]
  for node in root.iter(tag):
    txt=DOM_NodeText(node)
    if txt: vals.append(txt)
  return vals

#############################################################################
def DOM_GetLeafValsByTagName_minidom(root, tag):
  vals=[]
  for node in root.getElementsByTagName(tag):
    txt=DOM_NodeText(node)
    if txt: vals.append(txt)
  return vals

#############################################################################
def DOM_GetAttr(root, tag, attname):
  if type(root) in (ElementTree.Element, ElementTree.ElementTree):
    return DOM_GetAttr_ET(root, tag, attname)
  else:
    return DOM_GetAttr_minidom(root, tag, attname)

#############################################################################
def DOM_GetAttr_ET(root, tag, attname):
  vals=[]
  for node in root.iter(tag):
    if attname in node.attrib:
      vals.append(node.attrib[attname])
  return vals

#############################################################################
def DOM_GetAttr_minidom(root, tag, attname):
  vals=[]
  for node in root.getElementsByTagName(tag):
    if attname in node.attributes:
      vals.append(node.attributes[attname].value)
  return vals

#############################################################################
def DOM_GetNodeAttr(node, attname):
  if type(node) in (ElementTree.Element, ElementTree.ElementTree):
    return DOM_GetNodeAttr_ET(node, attname)
  else:
    return DOM_GetNodeAttr_minidom(node, attname)

#############################################################################
def DOM_GetNodeAttr_ET(node, attname):
  return node.attrib[attname] if attname in node.attrib else None

#############################################################################
def DOM_GetNodeAttr_minidom(node, attname):
  return node.attributes[attname].value if attname in node.attributes else None

#############################################################################
def XpathFind(xp, root):
  if type(root) in (ElementTree.Element, ElementTree.ElementTree):
    return root.findall(xp)
  else:
    import xpath #for minidom
    return xpath.find(xp, root)

#############################################################################
def Describe(fin):
  tags={}
  et_itrabl = ElementTree.iterparse(fin, events=("start", "end")) #iterable
  et_itr = iter(et_itrabl) #iterator
  event, root = et_itr.next() #root element
  n_elem=0; n_term=0;
  while True:
    try:
      ee = et_itr.next()
    except Exception as e:
      break
    event, elem = ee
    n_elem+=1
    if event == 'start':
      if elem.tag not in tags:
        tags[elem.tag] = 0
      tags[elem.tag] += 1
    logging.debug('event:%6s, elem.tag:%6s, elem.text:%6s'%(event, elem.tag, elem.text))
    for k,v in elem.attrib.items():
      logging.debug('\telem.attrib["%s"]: %s'%(k, v))
    n_term+=1
  logging.debug('n_elem: %d'%n_elem)
  logging.debug('n_term: %d'%n_term)
  for tag in sorted(tags.keys()):
    logging.info('%24s: %6d'%(tag, tags[tag]))

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='XML utility', epilog='clean: UTF-8 encoding compliant')

  ops = ['match_xpath', 'describe', 'clean']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input XML file")
  parser.add_argument("--xpath", help="xpath pattern")
  parser.add_argument("--force", type=bool, help="ignore UTF-8 encoding errors")
  parser.add_argument("--o", dest="ofile", help="output XML file")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  codecs_mode = "ignore" if args.force else "replace"

  fin = open(args.ifile, "r") if args.ifile else sys.stdin
  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  if args.op == "describe":
    Describe(fin)

  elif args.op == "clean":
    try:
      root = ElementTree.parse(fin)
    except Exception as e:
      parser.error('ElementTree.parse(): %s'%str(e))
    fin.close()
    root.write(fout, encoding="UTF-8", xml_declaration=True)

  elif args.op == "match_xpath":
    try:
      root = ElementTree.parse(fin)
    except Exception as e:
      parser.error('ElementTree.parse(): %s'%str(e))
    fin.close()
    if not (xp and ifile):
      parser.error('--i and --xpath required.')
    nodes = XpathFind(xp, root)
    fout.write('"xpath","match"\n')
    for i,node in enumerate(nodes):
      fout.write('"%s","%s"\n'%(xp, DOM_NodeText(node)))
    logging.info('matches: %d'%len(nodes))

  else:
    parser.error('Invalid operation. %s'%(args.op))



#!/usr/bin/env python3
###
import sys,os,re,json,time,logging
#
from ..util import rest_utils
#
##############################################################################
def ListStudies(base_url, fout):
  """Only simple metadata."""
  url_this=base_url+'/studies?size=100'
  tags=[]; n_study=0; rval=None;
  while True:
    if rval:
      if 'next' not in rval['_links']: break
      elif url_this == rval['_links']['last']['href']: break
      else: url_this = rval['_links']['next']['href']
    logging.debug(url_this)
    rval = rest_utils.GetURL(url_this, parse_json=True)
    if not rval or '_embedded' not in rval or 'studies' not in rval['_embedded']: break
    studies = rval['_embedded']['studies']
    if not studies: break
    for study in studies:
      if not tags:
        for tag in study.keys():
          if type(study[tag]) not in (list, dict): tags.append(tag) #Only simple metadata.
        fout.write('\t'.join(tags)+'\n')
      n_study+=1
      vals = [str(study[tag]).replace('\n', ' ') if tag in study and study[tag] is not None else '' for tag in tags]
      fout.write('\t'.join(vals)+'\n')
  logging.info('n_study: {0}'.format(n_study))

##############################################################################
def SearchStudies(base_url, ids, searchtype, fout):
  url=base_url+'/studies/search'
  if searchtype=='gcst':
    url+='/'
  elif searchtype.lower()=='pubmedid':
    url+='/findByPublicationIdPubmedId?pubmedId='
  elif searchtype.lower()=='efotrait':
    url+='/findByEfoTrait?efoTrait='
  elif searchtype.lower()=='efouri':
    url+='/findByEfoUri?efoUri='
  elif searchtype.lower()=='accessionid':
    url+='/findByAccessionId?accessionId='
  else:
    logging.error('Not yet supported: searchtype: %s'%(searchtype))
    return
  tags=[]; n_study=0; rval=None;
  for id_this in ids:
    url_this=url+'%s'%id_this
    rval = rest_utils.GetURL(url_this, parse_json=True)
    if not rval or '_embedded' not in rval or 'studies' not in rval['_embedded']: continue
    studies = rval['_embedded']['studies']
    if not studies: continue
    for study in studies:
      if not tags:
        for tag in study.keys():
          if type(study[tag]) not in (list, dict): tags.append(tag) #Only simple metadata.
        fout.write('\t'.join(tags)+'\n')
      n_study+=1
      vals = [str(study[tag]).replace('\n', ' ') if tag in study and study[tag] is not None else '' for tag in tags]
      fout.write('\t'.join(vals)+'\n')
    logging.debug('%s'%json.dumps(rval, sort_keys=True, indent=2))
  logging.info('n_study: {0}'.format(n_study))

##############################################################################
def GetStudyAssociations(base_url, ids, fout):
  """
	Mapped genes via SNP links.
	arg = authorReportedGene
	sra = strongestRiskAllele
	https://www.ebi.ac.uk/gwas/rest/api/studies/GCST002264/associations?projection=associationByStudy
  """
  url=base_url+'/studies'
  n_id=0; n_assn=0; n_loci=0; n_arg=0; n_sra=0; n_snp=0; gcsts=set([])
  tags_assn=[]; tags_study=[]; tags_locus=[]; tags_sra=[]; tags_arg=[];
  for id_this in ids:
    n_id+=1
    url_this=url+'/%s/associations?projection=associationByStudy'%id_this
    rval=rest_utils.GetURL(url_this, parse_json=True)
    if not rval:
      continue
    if '_embedded' in rval and 'associations' in rval['_embedded']:
      assns = rval['_embedded']['associations']
    else:
      logging.error('ERROR: no associations for study: %s'%id_this)
      continue

    for assn in assns:
      n_assn+=1
      if n_assn==1:
        for key,val in assn.items():
          if type(val) not in (list, dict):
            tags_assn.append(key)
        for key in assn['study'].keys():
          if type(assn['study'][key]) not in (list, dict):
            tags_study.append(key)
        for key in assn['loci'][0].keys():
          if key not in ('strongestRiskAlleles', 'authorReportedGenes'):
            tags_locus.append(key)
        for key in assn['loci'][0]['strongestRiskAlleles'][0].keys():
          if key != '_links':
            tags_sra.append(key)
#        for key in assn['loci'][0]['authorReportedGenes'][0].keys():
#          tags_arg.append(key)
        fout.write('\t'.join(list(map(lambda s: 'study_'+s, tags_study))
		+tags_assn
		+list(map(lambda s: 'locus_'+s, tags_locus))
#		+list(map(lambda s: 'reported_'+s, tags_arg))
		+list(map(lambda s: 'allele_'+s, tags_sra))
		+['snp_url'])+'\n')

      vals=[];
      for tag in tags_study:
        vals.append(CleanStr(assn['study'][tag]) if tag in assn['study'] and assn['study'][tag] is not None else '')
        try: gcsts.add(assn['study']['accessionId'])
        except: pass
      for tag in tags_assn:
        vals.append(CleanStr(assn[tag]) if tag in assn and assn[tag] is not None else '')
      for locus in assn['loci']:
        n_loci+=1
        vals_locus=[];
        for tag in tags_locus:
          vals_locus.append(CleanStr(locus[tag]) if tag in locus and locus[tag] is not None else '')

      ## Ignore reported genes.
#        for arg in locus['authorReportedGenes']:
#          n_arg+=1
#          vals_arg=[];
#          for tag in tags_arg:
#            if tag=='entrezGeneIds':
#              geneids = [g['entrezGeneId'] for g in arg[tag]]
#              if None in geneids: geneids.remove(None)
#              vals_arg.append(';'.join(geneids) if geneids else '')
#            elif tag=='ensemblGeneIds':
#              geneids = [g['ensemblGeneId'] for g in arg[tag]]
#              if None in geneids: geneids.remove(None)
#              vals_arg.append(';'.join(geneids) if geneids else '')
#            else:
#              vals_arg.append(CleanStr(arg[tag]) if tag in arg else '')
#          fout.write('\t'.join(vals+vals_locus+vals_arg+['' for tag in tags_sra]+[''])+'\n')

        for sra in locus['strongestRiskAlleles']:
          n_sra+=1
          vals_sra=[]; snp_href='';
          snp_href = sra['_links']['snp']['href'] if '_links' in sra and 'snp' in sra['_links'] and 'href' in sra['_links']['snp'] else ''
          if snp_href: n_snp+=1
          for tag in tags_sra:
            vals_sra.append(CleanStr(sra[tag]) if tag in sra and sra[tag] is not None else '')
#          fout.write('\t'.join(vals+vals_locus+['' for tag in tags_arg]+vals_sra+[snp_href])+'\n')
          fout.write('\t'.join(vals+vals_locus+vals_sra+[snp_href])+'\n')

#  logging.info('INPUT RCSTs: %d; OUTPUT RCSTs: %d ; assns: %d ; loci: %d ; reportedGenes: %d ; alleles: %d ; snps: %d'%(n_id, len(gcsts), n_assn, n_loci, n_arg, n_sra, n_snp))
  logging.info('INPUT RCSTs: %d; OUTPUT RCSTs: %d ; assns: %d ; loci: %d ; alleles: %d ; snps: %d'%(n_id, len(gcsts), n_assn, n_loci, n_sra, n_snp))

##############################################################################
def GetSnps(base_url, ids, fout):
  """
	loc = location
	gc = genomicContext
  """
  url=base_url+'/singleNucleotidePolymorphisms'
  n_snp=0; n_gc=0; n_gene=0; n_gcloc=0; n_loc=0;
  tags=[]; tags_loc=[]; tags_gc=[]; tags_gcloc=[];  tags_gene=[]; 
  for id_this in ids:
    n_snp+=1
    url_this=url+'/'+id_this
    snp=rest_utils.GetURL(url_this, parse_json=True)
    if not snp:
      continue
    if n_snp==1:
      for key,val in snp.items():
        if type(val) not in (list, dict):
          tags.append(key)
      #for key in snp['locations'][0].keys():
      #  if key != '_links':
      #    tags_loc.append(key)
      for key in snp['genomicContexts'][0].keys():
        if key not in ('gene', 'location', '_links'):
          tags_gc.append(key)
      for key in snp['genomicContexts'][0]['location'].keys():
        if key != '_links':
          tags_gcloc.append(key)
      for key in snp['genomicContexts'][0]['gene'].keys():
        if key != '_links':
          tags_gene.append(key)
      fout.write('\t'.join(tags
		+list(map(lambda s: 'genomicContext_'+s, tags_gc))
		+list(map(lambda s: 'loc_'+s, tags_gcloc))
		+list(map(lambda s: 'gene_'+s, tags_gene)))
		+'\n')
    vals=[];
    for tag in tags:
      vals.append(CleanStr(snp[tag]) if tag in snp and snp[tag] is not None else '')
    for gc in snp['genomicContexts']:
      n_gc+=1
      vals_gc=[];
      for tag in tags_gc:
        vals_gc.append(CleanStr(gc[tag]) if tag in gc and gc[tag] is not None else '')
      gcloc = gc['location']
      n_gcloc+=1
      vals_gcloc=[];
      for tag in tags_gcloc:
        if tag=='region':
          vals_gcloc.append(gcloc[tag]['name'] if tag in gcloc and gcloc[tag] and 'name' in gcloc[tag] and gcloc[tag]['name'] is not None else '')
        else:
          vals_gcloc.append(CleanStr(gcloc[tag]) if tag in gcloc and gcloc[tag] is not None else '')
      gene = gc['gene']
      n_gene+=1
      vals_gene=[];
      for tag in tags_gene:
        if tag=='entrezGeneIds':
          geneids = [g['entrezGeneId'] for g in gene[tag]]
          if None in geneids: geneids.remove(None)
          vals_gene.append(';'.join(geneids) if geneids else '')
        elif tag=='ensemblGeneIds':
          geneids = [g['ensemblGeneId'] for g in gene[tag]]
          if None in geneids: geneids.remove(None)
          vals_gene.append(';'.join(geneids) if geneids else '')
        else:
          vals_gene.append(str(gene[tag]) if tag in gene and gene[tag] is not None else '')
      fout.write('\t'.join(vals+vals_gc+vals_gcloc+vals_gene)+'\n')

  logging.info('SNPs: %d ; genomicContexts: %d ; genes: %d ; locations: %d'%(n_snp, n_gc, n_gene, n_gcloc))

##############################################################################
def CleanStr(val):
  val = re.sub(r'[\t\r\n]', ' ', str(val))
  return val.strip()




#!/usr/bin/env python3
'''
	GeneOntology client.
'''
import os,sys,re,json,time,logging
import urllib.parse

from ..util import rest_utils

#
##############################################################################
def GetEntities(base_url, ids, fout): 
  """For only one type of entity per call (gene, term)."""
  n_ent=0; tags=[];
  for id_this in ids:
    ent = rest_utils.GetURL(base_url+'/bioentity/%s'%(urllib.parse.quote(id_this)), parse_json=True)
    logging.debug(json.dumps(ent, sort_keys=True, indent=2))
    n_ent+=1
    if not tags:
      tags = ent.keys()
      fout.write('%s\n'%('\t'.join(tags)))
    vals = [str(ent[tag]) if tag in ent else '' for tag in tags]
    fout.write('%s\n'%('\t'.join(vals)))
  logging.info('n_ent: %d'%(n_ent))

##############################################################################
def GetGeneTerms(base_url, ids, fout): 
  n_assn=0; tags=[];
  for id_this in ids:
    rval = rest_utils.GetURL(base_url+'/bioentity/gene/%s/function'%(urllib.parse.quote(id_this)), parse_json=True)
    assns = rval['associations'] if 'associations' in rval else []
    for assn in assns:
      logging.debug(json.dumps(assn, sort_keys=True, indent=2))
      n_assn+=1
      if not tags:
        tags = assn.keys()
        fout.write('%s\n'%('\t'.join(tags)))
      vals = [str(assn[tag]) if tag in assn else '' for tag in tags]
      fout.write('%s\n'%('\t'.join(vals)))
  logging.info('n_gene: %d; n_assn: %d'%(len(ids), n_assn))

##############################################################################



#!/usr/bin/env python3
"""
Reactome REST API utilities.
https://reactome.org/ContentService/
https://reactome.org/ContentService/data/pathways/top/9606
"""
import sys,os,re,json,time,logging
#
from ..util import rest_utils
#
#
OFMTS={'XML':'application/xml','JSON':'application/json'}
#
HEADERS={'Content-type':'text/plain','Accept':OFMTS['JSON']}
#
##############################################################################
def ListToplevelPathways(base_url, fout):
  n_out=0;
  species="9606";
  try:
    rval = rest_utils.GetURL(base_url+'/data/pathways/top/%s'%(species), parse_json=True)
  except Exception as e:
    logging.error('%s'%(e))
  pathways = rval
  tags = None;
  for pathway in pathways:
    if not tags:
      tags = pathway.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [str(pathway[tag]) if tag in pathway else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('Top-level pathways: %d'%n_out)

##############################################################################
def ListDiseases(base_url, fout):
  n_out=0;
  try:
    rval = rest_utils.GetURL(base_url+'/getDiseases')
  except Exception as e:
    logging.error('%s'%(e))
  tags=['identifier','dbId','displayName','definition','schemaClass']
  fout.write('\t'.join(tags)+'\n')
  for line in rval.splitlines():
    idval, doid = re.split(r'\t', line)
    disease = GetDisease(base_url, idval)
    vals=[]
    for tag in tags:
      vals.append(disease[tag] if tag in disease else '')
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('diseases: %d'%n_out)

##############################################################################
def ListProteins(base_url, fout):
  n_out=0;
  try:
    rval = rest_utils.GetURL(base_url+'/getUniProtRefSeqs')
  except Exception as e:
    logging.error('%s'%(e))
  tags=['identifier','dbId','species','name','displayName','description','definition','schemaClass']
  fout.write('\t'.join(tags)+'\n')
  for line in rval.splitlines():
    idval, uniprot = re.split(r'\t', line)
    protein = GetProtein(base_url, idval)
    vals=[]
    for tag in tags:
      vals.append(protein[tag] if tag in protein else '')
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('proteins: %d'%n_out)

##############################################################################
def ListCompounds(base_url, fout):
  try:
    from rdkit import Chem
  except Exception as e:
    logging.error("RDKit not installed.")
    return
  n_out=0;
  try:
    rval = rest_utils.GetURL(base_url+'/getReferenceMolecules')
  except Exception as e:
    logging.error('%s'%(e))
  tags=['identifier','dbId','name','displayName','formula','schemaClass']
  fout.write('\t'.join(tags+['smiles'])+'\n')
  for line in rval.splitlines():
    idval, chebi = re.split(r'\t', line)
    compound = GetCompound(base_url, idval)
    vals=[]
    if (type(compound['name']) is types.ListType):
      compound['name']= compound['name'][0]
    for tag in tags:
      vals.append(compound[tag] if tag in compound else '')
    mdlstr = compound['atomicConnectivity'] if (compound and 'atomicConnectivity' in compound) else ''
    smiles=''
    if mdlstr:
      #logging.info('DEBUG: mdlstr="%s"'%mdlstr)
      mdlstr = re.sub(r'\\n', '\n', mdlstr)
      try:
        mol = Chem.MolFromMolBlock(mdlstr)
        smiles = Chem.MolToSmiles(mol)
      except Exception as e:
        logging.error('RDKit Error: %s'%(e))
    fout.write('\t'.join(vals+[smiles])+'\n')
    n_out+=1
  logging.info('compounds: %d'%n_out)

##############################################################################
def GetDisease(base_url, did, fout):
  GetEntity(base_url, [did], 'Disease', fout)

##############################################################################
def GetProtein(base_url, pid, fout):
  GetEntity(base_url, [pid], 'EntityWithAccessionedSequence', fout)

##############################################################################
def GetCompound(base_url, cid, fout):
  GetEntity(base_url, [cid], 'ReferenceMolecule', fout)

##############################################################################
def GetEntity(base_url, ids, ctype, fout):
  for id_this in ids:
    rval=rest_utils.GetURL(base_url+'/queryById/%s/%s'%(ctype, id_this), parse_json=True)
    fout.write(json.dumps(rval, sort_keys=True, indent=2)+'\n')

##############################################################################
def PathwaysForEntities(base_url, ids, fout):
  n_all=0; n_out=0; n_err=0;
  d=('ID='+(','.join(ids))) ##plain text POST body, e.g. "ID=170075,176374,68557"

  rval=rest_utils.PostURL(base_url+'/pathwaysForEntities', data=d, headers=HEADERS, parse_json=True)
  #fout.write(json.dumps(rval, sort_keys=True, indent=2)+'\n')

  pathways = rval
  tags=[];
  for pathway in pathways:
    n_all+=1
    if n_all==1 or not tags:
      tags=sorted(list(set(pathway.keys()) - set(['species','stableIdentifier'])))
      fout.write(','.join(tags+['stableId','stableIdName'])+'\n')

    vals=[];
    for tag in tags:
      vals.append(pathway[tag] if tag in pathway else '')
    vals.append(pathway['stableIdentifier']['dbId'])
    vals.append(pathway['stableIdentifier']['displayName'])

    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all, n_out, n_err))

##############################################################################
def PathwaysForGenes(base_url, ids, fout):
  n_all=0; n_out=0; n_err=0;
  d=(','.join(ids)) ##plain text POST body

  rval=rest_utils.PostURL(base_url+'/queryHitPathways', data=d, headers=HEADERS, parse_json=True)
  #fout.write(json.dumps(rval, sort_keys=True, indent=2)+'\n')

  pathways = rval
  tags=[];
  for pathway in pathways:
    n_all+=1
    if n_all==1 or not tags:
      tags=sorted(list(set(pathway.keys()) - set(['species','stableIdentifier'])))
      fout.write('\t'.join(tags+['stableId','stableIdName'])+'\n')

    vals=[];
    for tag in tags:
      vals.append(pathway[tag] if tag in pathway else '')
    vals.append(pathway['stableIdentifier']['dbId'])
    vals.append(pathway['stableIdentifier']['displayName'])

    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all, n_out, n_err))

##############################################################################
def PathwayParticipants(base_url, id_query, fout):
  n_all=0; n_out=0; n_err=0;
  rval = rest_utils.GetURL(base_url+'/pathwayParticipants/%s'%(id_query), parse_json=True)
  fout.write(json.dumps(rval, sort_keys=True, indent=2)+'\n')

  if not (rval and type(rval) is types.ListType):
    return
  parts = rval
  tags=[];
  for part in parts:
    n_all+=1
    if n_all==1 or not tags:
      tags=sorted(list(set(part.keys()) - set(['species','stableIdentifier'])))
      fout.write('\t'.join(tags+['stableId','stableIdName'])+'\n')

    vals=[];
    for tag in tags:
      vals.append(part[tag] if tag in part else '')

    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all, n_out, n_err))

##############################################################################



#!/usr/bin/env python3
#############################################################################
### CMap Clue API client.
### CMap is the project; Clue is the platform.
### https://clue.io/api
### https://clue.io/connectopedia/query_api_tutorial
### https://clue.io/connectopedia/perturbagen_types_and_controls
### l1000_type: "landmark"|"inferred"|"best inferred"|"not inferred"
### BING = Best inferred plus Landmark
### API services: cells, genes, perts, pcls, plates, profiles, sigs,
### probeset_to_entrez_id, rep_fda_exclusivity
#############################################################################
### Apps:
### https://clue.io/cell-app
### https://clue.io/repurposing-app
#############################################################################
### curl -X GET --header "user_key: YOUR_KEY_HERE" 'https://api.clue.io/api/cells?filter=\{"where":\{"provider_name":"ATCC"\}\}' |python3 -m json.tool
#############################################################################
import sys,os,re,argparse,json,logging,yaml
import urllib.parse

from ..util import rest_utils

N_CHUNK=200

#############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

#############################################################################
def ListDatatypes(base_url, params):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url = (base_url+'/dataTypes')
  response = rest_utils.GetURL(url, headers=headers, parse_json=True)
  logging.debug(json.dumps(response, indent=2))

#############################################################################
def ListDatasets(base_url, params):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url = (base_url+'/datasets')
  response = rest_utils.GetURL(url, headers=headers, parse_json=True)
  logging.debug(json.dumps(response, indent=2))

#############################################################################
def ListPerturbagenClasses(base_url, params, fout):
  headers = {"user_key":params['user_key']}
  url = (base_url+'/pcls')
  pcls = rest_utils.GetURL(url, headers=headers, parse_json=True)
  tags=None; n_pcl=0;
  for pcl in pcls:
    n_pcl+=1
    if not tags:
      tags=pcl.keys()
      fout.write('\t'.join(tags)+'\n')
    vals=[];
    for tag in tags:
      if tag not in pcl:
        vals.append('')
      else:
        vals.append(str(pcl[tag]))
    fout.write('\t'.join(vals)+'\n')
  logging.info('pcls: %d'%n_pcl)

#############################################################################
def GetGenes(base_url, params, ids, id_type, fout):
  url_base=(base_url+'/genes?user_key='+params['user_key'])
  tags = None
  n_gene=0;
  for id_this in ids:
    n_gene_this=0;
    logging.info('id: %s'%id_this)
    i_chunk=0;
    while True:
      qry=('{"where":{"%s":"%s"},"skip":%d,"limit":%d}'%(
	id_type, urllib.parse.quote(id_this), i_chunk*N_CHUNK, N_CHUNK))
      url=url_base+('&filter=%s'%(qry))
      try:
        genes = rest_utils.GetURL(url, parse_json=True)
      except:
        break
      if not genes:
        break
      for gene in genes:
        n_gene_this+=1
        if not tags:
          tags = gene.keys()
          fout.write('\t'.join(tags)+'\n')
        vals = []
        for tag in tags:
          if tag not in gene:
            vals.append('')
          elif type(gene[tag]) in (list, tuple):
            vals.append(';'.join([str(x) for x in gene[tag]]))
          else:
            vals.append(str(gene[tag]))
        fout.write('\t'.join(vals)+'\n')
      i_chunk+=1
    logging.info('\tgenes: %d'%n_gene_this)
    n_gene+=n_gene_this
  logging.info('genes: %d'%n_gene)

#############################################################################
def ListGenes_Landmark(base_url, params, fout):
  GetGenes(base_url, params, ['landmark'], 'l1000_type', fout)

#############################################################################
def ListGenes(base_url, params, fout):
  GetGenes(base_url, params, ['landmark', 'inferred', 'best inferred', 'not inferred'], 'l1000_type', fout)

#############################################################################
### pert_type:
###	trt_cp - Compound
###	trt_sh - shRNA for loss of function (LoF) of gene
###	trt_lig - Peptides and other biological agents (e.g. cytokine)
###	trt_sh.cgs - Consensus signature from shRNAs targeting the same gene
#############################################################################
def GetPerturbagens(base_url, params, ids, id_type, fout):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/perts')
  tags = None
  fields = ['pert_id', 'pert_iname', 'pert_type', 'pert_vendor', 'pert_url', 'id', 'pubchem_cid', 'entrez_geneId', 'vector_id', 'clone_name', 'oligo_seq', 'description', 'target', 'structure_url', 'moa', 'pcl_membership', 'tas', 'num_sig', 'status']
  n_pert=0;
  for id_this in ids:
    n_pert_this=0;
    logging.info('id: %s'%id_this)
    i_chunk=0;
    while True:
      qry = ('{"where":{"%s":"%s"},"fields":[%s],"skip":%d,"limit":%d}'%(
	id_type, urllib.parse.quote(id_this),
	(','.join(['"%s"'%f for f in fields])),
	i_chunk*N_CHUNK, N_CHUNK))
      url = url_base+('?filter=%s'%(qry))
      try:
        perts = rest_utils.GetURL(url, headers=headers, parse_json=True)
      except:
        continue
      if not perts:
        break
      logging.debug(json.dumps(perts, indent=2))
      for pert in perts:
        n_pert_this+=1
        if not tags:
          tags = pert.keys()
          fout.write('\t'.join(tags)+'\n')
        vals = []
        for tag in tags:
          if tag not in pert:
            vals.append('')
          elif type(pert[tag]) in (list, tuple):
            vals.append(';'.join([str(x) for x in pert[tag]]))
          else:
            vals.append(str(pert[tag]))
        fout.write('\t'.join(vals)+'\n')
      i_chunk+=1
    logging.info('\tperturbagens: %d'%n_pert_this)
    n_pert+=n_pert_this
  logging.info('perturbagens: %d'%n_pert)

#############################################################################
def ListPerturbagens(base_url, params, fout):
  pert_types = ['trt_cp', 'trt_lig', 'trt_sh', 'trt_sh.cgs', 'trt_oe', 'trt_oe.mut', 'trt_xpr', 'trt_sh.css', 'ctl_vehicle.cns', 'ctl_vehicle', 'ctl_vector', 'ctl_vector.cns', 'ctl_untrt.cns', 'ctl_untrt']
  GetPerturbagens(base_url, params, pert_types, 'pert_type', fout)

#############################################################################
def ListDrugs(base_url, params, fout):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/rep_drugs')
  tags = None
  n_drug=0; i_chunk=0;
  while True:
    qry = ('{"skip":%d,"limit":%d}'%(i_chunk*N_CHUNK, N_CHUNK))
    url = url_base+('?filter=%s'%(qry))
    try:
      drugs = rest_utils.GetURL(url, headers=headers, parse_json=True)
    except Exception as e:
      logging.error('Exception: %s'%e)
      continue
    if not drugs:
      break
    for drug in drugs:
      n_drug+=1
      if not tags:
        tags = drug.keys()
        fout.write('\t'.join(tags)+'\n')
      vals = []
      for tag in tags:
        if tag not in drug:
          vals.append('')
        elif type(drug[tag]) in (list, tuple):
          vals.append(';'.join([str(x) for x in drug[tag]]))
        else:
          vals.append(str(drug[tag]))
      fout.write('\t'.join(vals)+'\n')
    i_chunk+=1
  logging.info('drugs: %d'%n_drug)

#############################################################################
def GetCells(base_url, params, ids, id_type, fout):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/cells')
  tags = None
  n_cell=0;
  for id_this in ids:
    n_cell_this=0;
    logging.info('id: %s'%id_this)
    i_chunk=0;
    while True:
      qry = ('{"where":{"%s":"%s"},"skip":%d,"limit":%d}'%(
	id_type, urllib.parse.quote(id_this),
	i_chunk*N_CHUNK, N_CHUNK))
      url = url_base+('?filter=%s'%(qry))
      try:
        cells = rest_utils.GetURL(url, headers=headers, parse_json=True)
      except:
        break
      if not cells:
        break
      logging.debug(json.dumps(cells, indent=2))
      for cell in cells:
        n_cell_this+=1
        if not tags:
          tags = cell.keys()
          fout.write('\t'.join(tags)+'\n')
        vals = []
        for tag in tags:
          if tag not in cell:
            vals.append('')
          elif type(cell[tag]) in (list, tuple):
            vals.append(';'.join([str(x) for x in cell[tag]]))
          else:
            vals.append(str(cell[tag]))
        fout.write('\t'.join(vals)+'\n')
      i_chunk+=1
    logging.info('\tcells: %d'%n_cell_this)
    n_cell+=n_cell_this
  logging.info('cells: %d'%n_cell)

#############################################################################
# 2570 (3/28/2019)
def ListCells(base_url, params, fout):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/cells')
  tags = None
  n_cell=0; i_chunk=0;
  while True:
    qry = ('{"skip":%d,"limit":%d}'%(i_chunk*N_CHUNK, N_CHUNK))
    url = url_base+('?filter=%s'%(qry))
    try:
      cells = rest_utils.GetURL(url, headers=headers, parse_json=True)
    except Exception as e:
      logging.error('Exception: %s'%e)
      continue
    if not cells:
      break
    for cell in cells:
      n_cell+=1
      if not tags:
        tags = cell.keys()
        fout.write('\t'.join(tags)+'\n')
      vals = []
      for tag in tags:
        if tag not in cell:
          vals.append('')
        elif type(cell[tag]) in (list, tuple):
          vals.append(';'.join([str(x) for x in cell[tag]]))
        else:
          vals.append(str(cell[tag]))
      fout.write('\t'.join(vals)+'\n')
    i_chunk+=1
  logging.info('cells: %d'%n_cell)

#############################################################################
def CountSignatures(base_url, params, args):
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/sigs/count')
  url = url_base+('?where=%s'%(args.clue_where))
  try:
    sigs = rest_utils.GetURL(url, headers=headers, parse_json=True)
    logging.info('signatures: %d'%len(sigs))
  except Exception as e:
    logging.error('%s'%e)

#############################################################################
def GetSignatures(base_url, params, args, fout):
  fields = ['pert_id','pert_iname','pert_desc','pert_dose','cell_id','provenance_code','target_seq','target_is_lm','target_is_bing','target_zs','dn100_bing','up100_bing']
  headers = {"Accept":"application/json", "user_key":params['user_key']}
  url_base = (base_url+'/sigs')
  tags=None; n_sig=0; i_chunk=0;
  while True:
    qry = ('{"where":%s,"fields":[%s],"skip":%d,"limit":%d}'%(
	args.clue_where, (','.join(['"%s"'%f for f in fields])),
	args.skip+i_chunk*N_CHUNK, N_CHUNK))
    url = url_base+('?filter=%s'%(qry))
    try:
      sigs = rest_utils.GetURL(url, headers=headers, parse_json=True)
    except:
      continue
    if not sigs:
      break
    logging.debug(json.dumps(sigs, indent=2))
    for sig in sigs:
      n_sig+=1
      if not tags:
        tags = sig.keys()
        fout.write('\t'.join(tags)+'\n')
      vals = []
      for tag in tags:
        if tag not in sig:
          vals.append('')
        elif type(sig[tag]) in (list, tuple):
          vals.append(';'.join([str(x) for x in sig[tag]]))
        else:
          vals.append(str(sig[tag]))
      fout.write('\t'.join(vals)+'\n')
      if n_sig>=args.nmax: break
    if n_sig>=args.nmax: break
    i_chunk+=1
  logging.info('signatures: %d'%n_sig)

#############################################################################
if __name__=="__main__":
  API_HOST = "api.clue.io"
  API_BASE_PATH = "/api"

  epilog='''\
CMap is the project; Clue is the platform.
https://clue.io/api. 
Credentials config file should be at $HOME/.clueapi.yaml.
'''
  parser = argparse.ArgumentParser(description='CLUE.IO REST API client utility', epilog=epilog)
  ops = ['getGenes', 'listGenes', 'listGenes_landmark', 
	'getPerturbagens', 'listPerturbagens', 'listDrugs',
	'countSignatures', 'getSignatures',
	'getCells', 'listCells',
	'listPerturbagenClasses',
	'listDatasets', 'listDatatypes',
	'search']
  id_types = ['cell_id', 'pert_id', 'gene_symbol', 'entrez_id']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", help="IDs, comma-separated")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--id_type", choices=id_types, help='query ID or field type, e.g. gene_symbol')
  parser.add_argument("--clue_where", help="Clue API search where, e.g. '{\"pert_desc\":\"sirolimus\",\"cell_id\":\"MCF7\"}'")
  parser.add_argument("--nmax", type=int, default=1000, help="max results")
  parser.add_argument("--skip", type=int, default=0, help="skip results")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.clueapi.yaml")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)

  args = parser.parse_args()

  params = ReadParamFile(args.param_file)

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  if args.ifile:
    fin = open(args.ifile)
    if not fin: parser.error('Cannot open: %s'%args.ifile)
    ids=[];
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
    logging.info('input queries: %d'%(len(ids)))
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  if args.op=='getGenes':
    if not ids: parser.error('--ids or --i required.')
    GetGenes(base_url, params, ids, args.id_type, fout)

  elif args.op=='listGenes':
    ListGenes(base_url, params, fout)

  elif args.op=='listGenes_landmark':
    ListGenes_Landmark(base_url, params, fout)

  elif args.op=='getPerturbagens':
    if not ids: parser.error('--ids or --i required.')
    GetPerturbagens(base_url, params, ids, args.id_type, fout)

  elif args.op=='listPerturbagens':
    ListPerturbagens(base_url, params, fout)

  elif args.op=='listDrugs':
    ListDrugs(base_url, params, fout)

  elif args.op=='getCells':
    if not ids: parser.error('--ids or --i required.')
    GetCells(base_url, params, ids, args.id_type, fout)

  elif args.op=='listCells':
    ListCells(base_url, params, fout)

  elif args.op=='getSignatures':
    if not args.clue_where: parser.error('--clue_where required.')
    GetSignatures(base_url, params, args, fout)

  elif args.op=='countSignatures':
    if not args.clue_where: parser.error('--clue_where required.')
    CountSignatures(base_url, params, args)

  elif args.op=='listDatasets':
    ListDatasets(base_url, params)

  elif args.op=='listDatatypes':
    ListDatatypes(base_url, params)

  elif args.op=='listPerturbagenClasses':
    ListPerturbagenClasses(base_url, params, fout)

  else:
    parser.error('Unsupported operation: %s'%args.op)
    parser.print_help()



#!/usr/bin/env python3
"""
DrugCentral db utility functions.
"""
import os,sys,re,logging,yaml
import pandas,pandas.io.sql
import psycopg2,psycopg2.extras

#############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

#############################################################################
def Connect(dbhost, dbport, dbname, dbusr, dbpw):
  """Connect to db; specify default cursor type DictCursor."""
  dsn = ("host='%s' port='%s' dbname='%s' user='%s' password='%s'"%(dbhost, dbport, dbname, dbusr, dbpw))
  dbcon = psycopg2.connect(dsn)
  dbcon.cursor_factory = psycopg2.extras.DictCursor
  return dbcon

#############################################################################
def Version(dbcon, dbschema="public", fout=None):
  sql = ("SELECT * FROM {0}.dbversion".format(dbschema))
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  cur = dbcon.cursor()
  cur.execute(sql)
  row = cur.fetchone()
  tags = list(row.keys())
  fout.write("\t".join(tags)+"\n")
  fout.write("\t".join([str(row[tag]) for tag in tags])+"\n")

#############################################################################
def MetaListdbs(dbcon, fout=None):
  """Pg meta-command: list dbs from pg_database."""
  sql = ("SELECT pg_database.datname, pg_shdescription.description FROM pg_database LEFT OUTER JOIN pg_shdescription on pg_shdescription.objoid = pg_database.oid WHERE pg_database.datname ~ '^drug'")
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    fout.write("\t".join([str(row[tag]) for tag in tags])+"\n")

#############################################################################
def Describe(dbcon, dbschema="public", fout=None):
  '''Describing the schema.'''
  if fout:
    cur = dbcon.cursor()
    cur.execute(("SELECT table_name FROM information_schema.tables WHERE table_schema = %s"), (dbschema,))
    fout.write("schema\ttable\tcolumn\tdatatype\n")
    for row in cur:
      cur2 = dbcon.cursor()
      cur2.execute(("SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = %s AND table_name = %s"), (dbschema, row[0]))
      for row2 in cur2:
        fout.write("\t".join([dbschema, row[0], row2[0], row2[1]])+"\n")
      cur2.close()
    cur.close()
  else:
    df_out=None;
    sql1 = ("SELECT table_name FROM information_schema.tables WHERE table_schema = '{}'".format(dbschema))
    df1 = pandas.io.sql.read_sql_query(sql1, dbcon)
    for tname in df1.table_name:
      sql2 = ("SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = '{}' AND table_name = '{}'".format(dbschema, tname))
      df_this = pandas.io.sql.read_sql_query(sql2, dbcon)
      df_this["schema"] = dbschema
      df_this["table"] = tname
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
    df_out = df_out[["schema", "table", "column_name", "data_type"]]
    return df_out

#############################################################################
def Counts(dbcon, dbschema="public", fout=None):
  '''Listing the table rowcounts.'''
  if fout:
    n_table=0; n_row=0;
    cur = dbcon.cursor()
    cur.execute(("SELECT table_name FROM information_schema.tables WHERE table_schema = %s ORDER BY table_name"), (dbschema,))
    fout.write("schema\ttable\trowcount\n")
    cur2 = dbcon.cursor()
    for row in cur:
      n_table+=1
      cur2.execute("SELECT COUNT(*) FROM {}.{}".format(dbschema, row[0]))
      row2 = cur2.fetchone()
      n_row+=row2[0]
      fout.write("\t".join([dbschema, row[0], str(row2[0])])+"\n")
    cur2.close()
    cur.close()
    logging.info("Totals: tables: {}; rows: {}".format(n_table, n_row))
  else:
    df_out=None;
    sql1 = ("SELECT table_name FROM information_schema.tables WHERE table_schema = '{}'".format(dbschema))
    df1 = pandas.io.sql.read_sql_query(sql1, dbcon)
    for tname in df1.table_name:
      sql2 = ("SELECT COUNT(*) AS rowcount FROM {}.{}".format(dbschema, tname))
      df_this = pandas.io.sql.read_sql_query(sql2, dbcon)
      df_this["schema"] = dbschema
      df_this["table"] = tname
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
    df_out = df_out[["schema", "table", "rowcount"]]
    return df_out

#############################################################################
def ListStructures(dbcon, dbschema="public", fout=None):
  sql = ("SELECT id,name,cas_reg_no,smiles,inchikey,inchi,cd_formula AS formula,cd_molweight AS molweight FROM {}.structures".format(dbschema))
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {}".format(n_out))

#############################################################################
def ListStructures2Smiles(dbcon, dbschema="public", fout=None):
  sql = ("SELECT smiles, id, name FROM {0}.structures WHERE smiles IS NOT NULL".format(dbschema))
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; 
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    vals = [row["smiles"], str(row["id"]), row["name"]]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListStructures2Molfile(dbcon, dbschema="public", fout=None):
  n_out=0;
  cur = dbcon.cursor()
  cur.execute("SELECT molfile, id, name FROM {0}.structures WHERE molfile IS NOT NULL".format(dbschema))
  for row in cur:
    molfile = re.sub(r'^[\s]*\n', str(row["id"])+"\n", row["molfile"])
    fout.write(molfile)
    fout.write("> <DRUGCENTRAL_STRUCT_ID>\n"+str(row["id"])+"\n\n")
    fout.write("> <NAME>\n"+row["name"]+"\n\n")
    fout.write("$$$$\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListProducts(dbcon, dbschema="public", fout=None):
  sql = ("SELECT * FROM {0}.product".format(dbschema))
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListActiveIngredients(dbcon, dbschema="public", fout=None):
  sql = ("SELECT * FROM {0}.active_ingredient".format(dbschema))
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListXrefTypes(dbcon, fout=None):
  sql="""\
SELECT DISTINCT
	id AS "dc_xref_type_id",
	type AS "dc_xref_type",
	description AS "dc_xref_type_description"
FROM
	id_type
ORDER BY
	dc_xref_type
"""
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {}".format(n_out))

#############################################################################
def ListIndications(dbcon, fout=None):
  sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	omop.umls_cui,
	omop.cui_semantic_type umls_semantic_type,
	omop.snomed_conceptid,
	omop.snomed_full_name
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
"""
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {}".format(n_out))

#############################################################################
def ListIndicationTargets(dbcon, fout=None):
  sql="""\
SELECT DISTINCT
	omop.concept_name AS omop_concept,
	omop.umls_cui,
	atf.struct_id,
	s.name AS struct_name,
	atf.target_id,
	atf.target_name,
	atf.gene,
	atf.action_type,
	atf.act_source,
	atf.act_type,
	atf.act_comment,
	atf.relation,
	atf.moa,
	atf.moa_source,
	atf.moa_source_url,
	r.pmid AS ref_pmid,
	r.doi AS ref_doi,
	r.title AS ref_title,
	r.dp_year AS ref_year
FROM
	act_table_full atf
	JOIN structures s ON s.id = atf.struct_id
	JOIN omop_relationship omop ON omop.struct_id = s.id
	LEFT OUTER JOIN reference r ON r.id = atf.moa_ref_id
WHERE
	omop.relationship_name = 'indication'
"""
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {}".format(n_out))

#############################################################################
def ListDrugdruginteractions(dbcon, fout=None):
  sql="""\
SELECT
	ddi.id AS ddi_id,
	ddi.drug_class1,
	ddi.drug_class2,
	ddi.source_id,
	drug_class1.id drug_class_id1,
	drug_class1.source source1,
	drug_class1.is_group is_group1,
	drug_class2.id drug_class_id2,
	drug_class2.source source2,
	drug_class2.is_group is_group2
FROM
	ddi
JOIN drug_class drug_class1 ON drug_class1.name = ddi.drug_class1
JOIN drug_class drug_class2 ON drug_class2.name = ddi.drug_class2
"""
  if not fout: return pandas.io.sql.read_sql_query(sql, dbcon)
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {}".format(n_out))

#############################################################################
def SearchIndications(dbcon, terms, fout=None):
  """Search names via Pg regular expression (SIMILAR TO)."""
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  for term in terms:
    sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	omop.umls_cui,
	omop.cui_semantic_type umls_semantic_type,
	omop.snomed_conceptid,
	omop.snomed_full_name
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
	AND (omop.concept_name ~* '{0}' OR omop.snomed_full_name ~* '{0}')
""".format(term)
    if fout:
      cur.execute(sql)
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql, dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])

  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetIndicationStructures(dbcon, ids, fout=None):
  """Input OMOP conceptIds."""
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	s.inchi,
	s.cd_formula,
	s.cd_molweight
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
	AND omop.concept_id = '{}'
"""
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructure(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql = ("""SELECT id,name,cas_reg_no,smiles,inchikey,inchi,cd_formula AS formula,cd_molweight AS molweight FROM structures WHERE id = '{}'""")
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureByXref(dbcon, xref_type, ids, fout=None):
  if not xref_type:
    logging.error("xref_type required.")
    return None
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql = ("""SELECT idn.identifier xref, idn.id_type xref_type, s.id dc_struct_id, s.name dc_struct_name FROM structures AS s JOIN identifier AS idn ON idn.struct_id=s.id WHERE idn.id_type = '"""+xref_type+"""' AND idn.identifier = '{}'""")
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureBySynonym(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql = ("""SELECT str.id, str.name structure_name, syn.name synonym FROM structures AS str JOIN synonyms AS syn ON syn.id=str.id WHERE syn.name = '{}'""")
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureIds(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql = ("""SELECT struct_id, id_type, identifier AS id FROM identifier WHERE struct_id = '{}'""")
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureProducts(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	p.id product_id,
	p.ndc_product_code,
	p.form product_form,
	p.generic_name product_generic_name,
	p.product_name,
	p.route product_route,
	p.marketing_status product_marketing_status,
	p.active_ingredient_count product_active_ingredient_count
FROM
	structures AS s
JOIN 
	active_ingredient ai ON ai.struct_id = s.id
JOIN 
	product p ON p.ndc_product_code = ai.ndc_product_code
WHERE 
	s.id = '{}'
"""
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureOBProducts(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	ob.id ob_id,
	ob.product_no ob_product_no,
	ob.ingredient,
	ob.dose_form,
	ob.route,
	ob.strength,
	ob.appl_type,
	ob.appl_no
FROM
	structures s
	JOIN struct2obprod s2ob ON s2ob.struct_id = s.id
	JOIN ob_product ob ON ob.id = s2ob.prod_id
WHERE 
	s.id = '{}'
"""
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetStructureAtcs(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	atc.code AS atc_code,
	atc.l1_code AS atc_l1_code,
	atc.l1_name AS atc_l1_name,
	atc.l2_code AS atc_l2_code,
	atc.l2_name AS atc_l2_name,
	atc.l3_code AS atc_l3_code,
	atc.l3_name AS atc_l3_name,
	atc.l4_code AS atc_l4_code,
	atc.l4_name AS atc_l4_name,
	atc.chemical_substance AS atc_substance
FROM
	structures AS s
JOIN 
	struct2atc ON struct2atc.struct_id = s.id
JOIN 
	atc ON atc.id = struct2atc.id
WHERE 
	s.id = '{}'
"""
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################
def GetProductStructures(dbcon, ids, fout=None):
  n_out=0; tags=None; df_out=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	p.id product_id,
	p.product_name,
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	s.inchi,
	s.cd_formula,
	s.cd_molweight
FROM
	product p
JOIN 
	active_ingredient ai ON p.ndc_product_code = ai.ndc_product_code
JOIN 
	structures s ON ai.struct_id = s.id
WHERE 
	p.id = '{}'
"""
  for id_this in ids:
    if fout:
      cur.execute(sql.format(id_this))
      for row in cur:
        if not tags:
          tags = list(row.keys())
          fout.write("\t".join(tags)+"\n")
        vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
        fout.write("\t".join(vals)+"\n")
        n_out+=1
    else:
      df_this = pandas.io.sql.read_sql_query(sql.format(id_this), dbcon)
      df_out = df_this if df_out is None else pandas.concat([df_out, df_this])
  if fout: logging.info("n_out: {}".format(n_out))
  else: return df_out

#############################################################################



#!/usr/bin/env python3
"""
Utility app for the PubChem PUG REST API.
"""
###
import sys,os,re,argparse,time,logging
#
from .. import pubchem
#
API_HOST='pubchem.ncbi.nlm.nih.gov'
API_BASE_PATH='/rest/pug'
#
##############################################################################
if __name__=='__main__':
  ops = [
        "list_sources_substance", "list_sources_assay",
        "get_name2sid", "get_name2cid", "get_name2synonyms",
        "get_smi2cid",
        "get_cid2smi", "get_cid2smiles", "get_cid2sdf",
        "get_cid2properties", "get_cid2inchi",
        "get_cid2synonyms", "get_cid2sid", "get_cid2assaysummary",
        "get_sid2cid", "get_sid2sdf", "get_sid2assaysummary",
        "get_assayname", "get_assaydescriptions", "get_assayresults" ]
  parser = argparse.ArgumentParser(description="PubChem PUG REST client")
  parser.add_argument("op",choices=ops,help='operation')
  parser.add_argument("--i", dest="ifile", help="input IDs file (CID|SID|SMILES|name)")
  parser.add_argument("--ids", help="input IDs (CID|SID|SMILES|name) (comma-separated)")
  parser.add_argument("--aids", help="input AIDs (comma-separated)")
  parser.add_argument("--iaid", dest="ifile_aid", help="input AIDs file")
  parser.add_argument("--o", dest="ofile", help="output (usually TSV)")
  parser.add_argument("--isomeric", action="store_true", help="return Isomeric SMILES")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=0)
  parser.add_argument("--nmax_per_cid", type=int, default=20)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  logging.info('Input IDs: %d'%(len(ids)))

  aids=[]
  if args.ifile_aid:
    fin = open(args.ifile_aid)
    while True:
      line = fin.readline()
      if not line: break
      aids.append(line.rstrip())
    fin.close()
  elif args.aids:
    aids = re.split(r'[,\s]+', args.aids)
  logging.info('Input AIDs: %d'%(len(aids)))

  t0=time.time()

  if args.op == 'list_sources_assay':
    pubchem.Utils.ListSources(BASE_URL, "assay", fout)

  elif args.op == 'list_sources_substance':
    pubchem.Utils.ListSources(BASE_URL, "substance", fout)

  elif args.op == 'get_cid2synonyms':
    pubchem.Utils.GetCID2Synonyms(BASE_URL, ids, args.skip, args.nmax, args.nmax_per_cid, fout)

  elif args.op == 'get_cid2properties':
    pubchem.Utils.GetCID2Properties(BASE_URL, ids, fout)

  elif args.op == 'get_cid2inchi':
    pubchem.Utils.GetCID2Inchi(BASE_URL, ids, fout)

  elif args.op == 'get_cid2sid':
    pubchem.Utils.GetCID2SID(BASE_URL, ids, fout)

  elif args.op == 'get_cid2smiles':
    pubchem.Utils.GetCID2Smiles(BASE_URL, ids, args.isomeric, fout)

  elif args.op == 'get_cid2sdf':
    pubchem.Utils.GetCID2SDF(BASE_URL, ids, fout)

  elif args.op == 'get_cid2assaysummary':
    pubchem.Utils.GetCID2AssaySummary(BASE_URL, ids, fout)

  elif args.op == 'get_sid2cid':
    pubchem.Utils.GetSID2CID(BASE_URL, ids, fout)

  elif args.op == 'get_sid2assaysummary':
    pubchem.Utils.GetSID2AssaySummary(BASE_URL, ids, fout)

  elif args.op == 'get_sid2sdf':
    pubchem.Utils.GetSID2SDF(BASE_URL, ids, fout, args.skip, args.nmax)

  elif args.op == 'get_smi2cid':
    pubchem.Utils.GetSmiles2CID(BASE_URL, ids, fout)

  elif args.op == 'get_name2sid':
    pubchem.Utils.GetName2SID(BASE_URL, ids, fout)

  elif args.op == 'get_name2cid':
    pubchem.Utils.GetName2CID(BASE_URL, ids, fout)

  elif args.op == 'get_name2synonyms':
    pubchem.Utils.GetName2Synonyms(BASE_URL, ids, fout)

  elif args.op == 'get_assayname':
    pubchem.Utils.GetAssayName(BASE_URL, aids, fout)

  elif args.op == 'get_assaydescriptions':
    pubchem.Utils.GetAssayDescriptions(BASE_URL, aids, args.skip, args.nmax, fout)

  elif args.op == 'get_assayresults':
    if not (aids and ids): parser.error('Input AIDs and SIDs required.')
    pubchem.Utils.GetAssaySIDResults(BASE_URL, aids, ids, args.skip, args.nmax, fout)

  else:
    parser.error('Invalid operation: %s'%args.op)

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))




#!/usr/bin/env python3
"""
LINCS REST API client
New (2019) iLINCS: 
http://www.ilincs.org/ilincs/APIinfo
http://www.ilincs.org/ilincs/APIdocumentation
(http://lincsportal.ccs.miami.edu/dcic/api/ DEPRECATED?)
"""
###
import sys,os,argparse,re,time,json,logging
#
from .. import lincs
#
#############################################################################
def GetGene(base_url, ids, fout):
  tags=None;
  for id_this in ids:
    url = base_url+'/GeneInfos/%s'%id_this
    rval = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2))
    if not tags:
      tags = [tag for tag in rval.keys() if type(rval[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(rval[tag]) if tag in rval else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("IDs: {0}".format(len(ids)))

#############################################################################
def GetDataset(base_url, ids, fout):
  tags=None;
  for id_this in ids:
    url = base_url+'/PublicDatasets/%s'%id_this
    rval = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2))
    if not tags:
      tags = [tag for tag in rval.keys() if type(rval[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(rval[tag]) if tag in rval else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("IDs: {0}".format(len(ids)))

#############################################################################
def GetCompound(base_url, ids, fout):
  tags=None; n_cpd=0;
  for id_this in ids:
    url = base_url+'/Compounds/%s'%id_this
    cpd = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(cpd, indent=2))
    if not tags:
      tags = [tag for tag in cpd.keys() if type(cpd[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(cpd[tag]) if tag in cpd else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_cpd+=1
  logging.info("IDs: {0}; n_cpd: {1}".format(len(ids), n_cpd))

#############################################################################
def SearchDataset(base_url, searchTerm, lincs, fout):
  tags=None;
  url = base_url+'/PublicDatasets/findTermMeta'
  d = {'term':searchTerm}
  if lincs: d['lincs'] = True
  rval = rest_utils.PostURL(url, data=d, parse_json=True)
  logging.debug(json.dumps(rval, indent=2))
  dsets = rval['data'] if 'data' in rval else []
  for dset in dsets:
    logging.debug(json.dumps(dset, indent=2))
    if not tags:
      tags = [tag for tag in dset.keys() if type(dset[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(dset[tag]) if tag in dset else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
  logging.info("Datasets: {0}".format(len(dsets)))

#############################################################################
def SearchSignature(base_url, ids, lincs, fout):
  #SignatureMeta?filter={"where":{"lincspertid":"LSM-2121"},"limit":10}
  tags=None; n_sig=0;
  for id_this in ids:
    url = base_url+'/SignatureMeta?filter={"where":{"lincspertid":"%s"}}'%id_this
    sigs = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(sigs, indent=2))
    for sig in sigs:
      logging.debug(json.dumps(sig, indent=2))
      if not tags:
        tags = [tag for tag in sig.keys() if type(sig[tag]) not in (list, dict)]
        fout.write('\t'.join(tags)+'\n')
      vals = [str(sig[tag]) if tag in sig else '' for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_sig+=1
  logging.info("IDs: {0}; n_sig: {1}".format(len(ids), n_sig))

#############################################################################
def GetSignature(base_url, ids, ngene, fout):
  tags=None; n_gene=0;
  url = base_url+'/ilincsR/downloadSignature'
  d = {'sigID':(','.join(ids)), 'display':True, 'noOfTopGenes':ngene}
  rval = rest_utils.PostURL(url, data=d, parse_json=True)
  logging.debug(json.dumps(rval, indent=2))
  genes = rval['data']['signature'] if 'data' in rval and 'signature' in rval['data'] else []
  for gene in genes:
    logging.debug(json.dumps(gene, indent=2))
    if not tags:
      tags = [tag for tag in gene.keys() if type(gene[tag]) not in (list, dict)]
      fout.write('\t'.join(tags)+'\n')
    vals = [str(gene[tag]) if tag in gene else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_gene+=1
  logging.info("IDs: {0}; n_gene: {1}".format(len(ids), n_gene))

#############################################################################
if __name__=='__main__':
  API_HOST="www.ilincs.org"
  API_BASE_PATH="/api"
  epilog="""\
Examples:
NCBI Gene IDs: 207;
PerturbagenIDs: BRD-A00100033 (get_compound);
LINCS PertIDs: LSM-2121;
Perturbagen-Compound IDs: LSM-2421;
Signature IDs: LINCSCP_10260,LINCSCP_10261,LINCSCP_10262;
Dataset IDs: EDS-1013,EDS-1014;
Search Terms: cancer, vorinostat, MCF7.
"""
  parser = argparse.ArgumentParser(description='LINCS REST API client (%s)'%(API_HOST), epilog=epilog)
  ops = ['get_gene', 'get_compound', 'get_dataset',
	'search_dataset', 'search_signature',
	'get_signature'
	]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--ids", help="input IDs, comma-separated")
  parser.add_argument("--searchTerm", dest="searchTerm", help="Entity searchTerm e.g. Rock1)")
  parser.add_argument("--lincs", action="store_true", help="LINCS datasets only")
  parser.add_argument("--ngene", type=int, default=50, help="top genes per signature")
  parser.add_argument("--nmax", type=int, help="max results")
  parser.add_argument("--skip", type=int, help="skip results")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url =' http://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split('[,\s]+', args.ids.strip())

  if args.op == 'get_gene':
    lincs.Utils.GetGene(base_url, ids, fout)

  elif args.op == 'get_compound':
    lincs.Utils.GetCompound(base_url, ids, fout)

  elif args.op == 'get_dataset':
    lincs.Utils.GetDataset(base_url, ids, fout)

  elif args.op == 'search_datasets':
    lincs.Utils.SearchDataset(base_url, args.searchTerm, args.lincs, fout)

  elif args.op == 'search_signatures':
    lincs.Utils.SearchSignature(base_url, ids, args.lincs, fout)

  elif args.op == 'get_signature':
    lincs.Utils.GetSignature(base_url, ids, args.ngene, fout)

  else:
    parser.error('Invalid operation: %s'%args.op)

